from __future__ import absolute_import
from pyramid.view import view_config
import tangos
from tangos import core
import numpy as np
from .halo_data import format_number, _relative_description
import sqlalchemy, sqlalchemy.orm
from six.moves import zip
from . import halo_from_request

class TimestepInfo(object):
    def __init__(self, ts):
        self.z = "%.2f"%ts.redshift
        self.t = "%.2e Gyr"%ts.time_gyr


class TimeLinks(object):
    def __init__(self, request, halo):
        link_names = ['earliest', '-10', '-1', '+1', '+10', 'latest']
        route_names = ['halo_earlier']*3 + ['halo_later']*3
        ns = ['inf',10,1,1,10,'inf']

        urls = [
            request.route_url(r, simid=halo.timestep.simulation.basename,
                              timestepid=halo.timestep.extension,
                              halonumber=halo.basename,
                              n=n)
            for r,n in zip(route_names, ns)
            ]

        self.urls = urls
        self.names = link_names

class DisplayProperty(object):
    def __init__(self, property):
        self.name = property.name.text
        self.value = format_property_data(property)
        self.is_array = property.data_is_array()

class TimeProperty(DisplayProperty):
    def __init__(self, halo):
        self.name = "t()"
        self.value = format_number(halo.timestep.time_gyr) + " Gyr"
        self.is_array = False

class RedshiftProperty(DisplayProperty):
    def __init__(self, halo):
        self.name = "z()"
        self.value = format_number(halo.timestep.redshift)
        self.is_array = False

def default_properties(halo):
    properties = [TimeProperty(halo), RedshiftProperty(halo)]

    for property in halo.properties.options(sqlalchemy.orm.joinedload(core.HaloProperty.name)):
        properties.append(DisplayProperty(property))

    return properties

def format_property_data(property):
    if property.data_is_array():
        """
        data = property.data_raw
        if len(data)>5 or len(data.shape)>1:
            return "size "+(" x ".join([str(s) for s in data.shape]))+" array"
        else:
            return "["+(",".join([_number_format(d) for d in data]))+"]"
        """
        return "Array"
    else:
        return format_number(property.data)

class SimulationInfo(object):
    def __init__(self, sim, request):
        self.name = sim.basename
        self.url = request.route_url('halo_in',simid=request.matchdict['simid'],
                                            timestepid=request.matchdict['timestepid'],
                                            halonumber=request.matchdict['halonumber'],
                                            n=sim.basename)


class HaloLinkInfo(object):
    def __init__(self, link, request):
        halo_source = link.halo_from
        halo_dest = link.halo_to
        weight_text = "( %.2f)"%link.weight if link.weight else ""
        self.name = "%s%s: %s"%(link.relation.text,weight_text,_relative_description(halo_source, halo_dest))
        self.url = request.route_url('halo_view', simid=halo_dest.timestep.simulation.basename,
                                     timestepid=halo_dest.timestep.extension,
                                     halonumber=halo_dest.basename)

def all_simulations(request):
    return [SimulationInfo(x,request) for x in tangos.all_simulations(request.dbsession)]

def halo_links(halo, request):
    links = []
    links_query = request.dbsession.query(core.HaloLink).filter_by(halo_from_id=halo.id).\
            order_by(core.HaloLink.weight.desc()).\
            options(sqlalchemy.orm.joinedload(core.HaloLink.halo_to).joinedload(core.Halo.timestep).joinedload(core.TimeStep.simulation))

    for lk in links_query.all():
        links.append(HaloLinkInfo(lk, request))
    return links

@view_config(route_name='halo_view', renderer='../templates/halo_view.jinja2')
def halo_view(request):
    halo = halo_from_request(request)
    ts = halo.timestep
    sim = ts.simulation

    return {'ts_info': TimestepInfo(ts),
            'this_id': halo.id,
            'halo_number': halo.halo_number,
            'halo_typetag': halo.tag,
            'timestep': ts.extension,
            'simulation': sim.basename,
            'all_simulations': all_simulations(request),
            'halo_links': halo_links(halo, request),
            'time_links': TimeLinks(request, halo),
            'properties': default_properties(halo),
            'halo_path': halo.path,
            'finder_id': halo.finder_id,
            'calculate_url': request.route_url('get_property',simid=request.matchdict['simid'],
                                            timestepid=request.matchdict['timestepid'],
                                            halonumber=request.matchdict['halonumber'],
                                            nameid="")[:-5],
            'tree_url': request.route_url('merger_tree',simid=request.matchdict['simid'],
                                            timestepid=request.matchdict['timestepid'],
                                            halonumber=request.matchdict['halonumber']),
            'gather_url': "/%s/%s/"%(sim.basename,ts.extension),
            'cascade_url': "/%s/%s/%s/"%(sim.basename,ts.extension,halo.basename)}


from .one_hop import HopStrategy, HopMajorDescendantStrategy, HopMajorProgenitorStrategy
from .multi_hop import MultiHopStrategy
from .multi_source import MultiSourceMultiHopStrategy
from .multi_hop_variants import MultiHopMajorDescendantsStrategy, MultiHopAllProgenitorsStrategy, \
    MultiHopMajorProgenitorsStrategy, MultiHopMostRecentMergerStrategy







from __future__ import absolute_import
from sqlalchemy import Index, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, clear_mappers
from .. import config

_verbose = False
_internal_session=None
_engine=None
Session=None

def get_default_session():
    """Get the default ORM session to be used when no other is specified.

    :rtype: sqlalchemy.orm.Session"""
    global _internal_session
    return _internal_session

def set_default_session(session):
    """Set the default ORM session to be used.

    No validation is carried out. If an inappropriate session is passed in, many database functions
    will subsequently fail. """
    global _internal_session
    _internal_session = session

def get_default_engine():
    """Get the default sqlalchemy engine to be used when no other is specified."""
    global _engine
    return _engine



clear_mappers()  # remove existing maps

Base = declarative_base()


from .dictionary import DictionaryItem
from .creator import Creator
from .simulation import Simulation, SimulationProperty
from .tracking import TrackData, update_tracker_halos
from .timestep import TimeStep
from .halo import Halo
from .halo_data import HaloProperty, HaloLink
from .stored_options import ArrayPlotOptions

Index("halo_index", HaloProperty.__table__.c.halo_id)
Index("name_halo_index", HaloProperty.__table__.c.name_id,
      HaloProperty.__table__.c.halo_id)
Index("halo_timestep_index", Halo.__table__.c.timestep_id)
Index("halo_creator_index", Halo.__table__.c.creator_id)
Index("halo_finder_index", Halo.__table__.c.finder_id)
Index("haloproperties_creator_index", HaloProperty.__table__.c.creator_id)
Index("halolink_index", HaloLink.__table__.c.halo_from_id)
Index("named_halolink_index", HaloLink.__table__.c.relation_id, HaloLink.__table__.c.halo_from_id)




def sim_query_from_name_list(names, session=None):
    if session is None:
        session = get_default_session()

    query = session.query(Simulation)
    clause = None
    if names is not None:
        for rxi in names:
            if clause is None:
                clause = (Simulation.basename == rxi)
            else:
                clause |= (Simulation.basename == rxi)

        query = query.filter(clause)

    return query

def sim_query_from_args(argv, session=None):
    if session is None:
        session = get_default_session()

    query = session.query(Simulation)

    if "for" in argv:
        rx = argv[argv.index("for") + 1:]
        clause = None
        for rxi in rx:
            if clause is None:
                clause = (Simulation.basename == rxi)
            else:
                clause = clause | (Simulation.basename == rxi)

        query = query.filter(clause)

    return query


def supplement_argparser(argparser):
    argparser.add_argument("--db-filename", help="Specify path to a database file to be used",
                           action='store', type=str, metavar="database_file.sqlite3")
    argparser.add_argument("--db-verbose", action='store_true',
                           help="Switch on sqlalchemy echo mode")


def process_options(argparser_options):
    global _verbose
    if argparser_options.db_filename is not None:
        config.db = argparser_options.db_filename
    _verbose = argparser_options.db_verbose


def init_db(db_uri=None, timeout=30, verbose=None):
    global _verbose, _internal_session, _engine, Session
    if db_uri is None:
        db_uri = 'sqlite:///' + config.db
    _engine = create_engine(db_uri, echo=verbose or _verbose,
                            isolation_level='READ UNCOMMITTED', connect_args={'timeout': timeout})
    Session = sessionmaker(bind=_engine)
    _internal_session=Session()
    Base.metadata.create_all(_engine)
    creator.set_creator(None)


from .dictionary import _get_dict_cache_for_session, get_dict_id, get_or_create_dictionary_item



__all__ = ['DictionaryItem',
           'sim_query_from_name_list', 'sim_query_from_args',
           'supplement_argparser',
           'update_tracker_halos',
           'process_options', 'init_db', 'Base', 'Creator',
           'get_default_session']



from __future__ import absolute_import
import pyparsing as pp
import functools

from . import StoredProperty, LiveProperty, FixedNumericInput, \
    FixedInput, Link, MultiCalculation, Calculation

def pack_args(for_function):
    """Return a version of for_function that takes a single argument instead of multiple arguments"""
    return lambda t: for_function(*t)


pp.ParserElement.enablePackrat()

numerical_value = pp.Regex(r'-?\d+(\.\d*)?([eE]-?\d+)?').setParseAction(pack_args(FixedNumericInput))

IN_OPS = [("**", "power"),
          ("*", "multiply"),
          ("/", "divide"),
          ("+", "add"),
          ("-", "subtract"),
          (">", "greater"),
          ("<", "less"),
          ("|", "logical_or"),
          ("&", "logical_and")]

UNARY_OPS = [("!", "logical_not")]

IN_OPS_PYPARSING = []
UNARY_OPS_PYPARSING = []

def generate_property_from_inop(opFunctionName, tokens):
    return LiveProperty(opFunctionName, *tokens[0])

for opSymbol, opFunctionName in IN_OPS:
    opGeneration = functools.partial(generate_property_from_inop, opFunctionName)
    IN_OPS_PYPARSING.append((pp.Literal(opSymbol).suppress(), 2, pp.opAssoc.RIGHT, opGeneration))

for opSymbol, opFunctionName in UNARY_OPS:
    opGeneration = functools.partial(generate_property_from_inop, opFunctionName)
    UNARY_OPS_PYPARSING.append((pp.Literal(opSymbol).suppress(), 1, pp.opAssoc.RIGHT, opGeneration))

property_name = pp.Word(pp.alphas,pp.alphanums+"_")
stored_property = property_name.setParseAction(pack_args(StoredProperty))

live_calculation_property = pp.Forward().setParseAction(pack_args(LiveProperty))

array_element = pp.Forward().setParseAction(pack_args(functools.partial(LiveProperty,"element")))


dbl_quotes = pp.Literal("\"").suppress()
sng_quotes = pp.Literal("'").suppress()

string_value = dbl_quotes.suppress() + pp.SkipTo(dbl_quotes).setParseAction(pack_args(FixedInput)) + dbl_quotes.suppress() | \
               sng_quotes.suppress() + pp.SkipTo(sng_quotes).setParseAction(pack_args(FixedInput)) + sng_quotes.suppress()

redirection = pp.Forward().setParseAction(pack_args(Link))


element_identifier = pp.Literal("[").suppress()+numerical_value+pp.Literal("]").suppress();

multiple_properties = pp.Forward().setParseAction(pack_args(MultiCalculation))

property_identifier = (redirection | array_element | live_calculation_property | stored_property | multiple_properties)


infix_operations = pp.infixNotation(numerical_value | property_identifier, IN_OPS_PYPARSING + UNARY_OPS_PYPARSING)


value_or_property_name = infix_operations | string_value | numerical_value |  property_identifier

multiple_properties << (pp.Literal("(").suppress()+value_or_property_name+pp.ZeroOrMore(pp.Literal(",").suppress()+value_or_property_name) +pp.Literal(")").suppress())

redirection << (live_calculation_property | stored_property ) + pp.Literal(".").suppress() + property_identifier

parameters = pp.Literal("(").suppress()+pp.Optional(value_or_property_name+pp.ZeroOrMore(pp.Literal(",").suppress()+value_or_property_name))+pp.Literal(")").suppress()
live_calculation_property << property_name+parameters

array_element << ((live_calculation_property | stored_property) + element_identifier)

property_complete = pp.stringStart()+value_or_property_name+pp.stringEnd()


def parse_property_name( name):
    return property_complete.parseString(name)[0]

def parse_property_name_if_required(name):
    if isinstance(name, Calculation):
        return name
    else:
        return parse_property_name(name)

def parse_property_names(*names):
    return MultiCalculation(*[parse_property_name(n) for n in names])

__all__ = ["parse_property_name", "parse_property_name_if_required", "parse_property_names"]


from .. import LiveHaloProperties
import numpy as np

class BHGal(LiveHaloProperties):
    def __init__(self, simulation=None, choose='BH_mass', minmax='max', bhtype='BH_central'):
        super(BHGal, self).__init__(simulation)
        self._maxmin = minmax
        self._choicep = choose
        self._bhtype = bhtype

    names = 'bh'

    def requires_property(self):
        return self._bhtype, self._bhtype+"."+self._choicep

    def live_calculate(self, halo, *args):
        if halo.object_typecode != 0:
            return None

        if self._bhtype not in list(halo.keys()):
            return None

        if type(halo[self._bhtype]) is list:
            chp = [bh[self._choicep] for bh in halo[self._bhtype] if self._choicep in bh]
            if len(chp) == 0:
                return None
            target = None
            if self._maxmin == 'min':
                target = np.argmin(chp)
            if self._maxmin == 'max':
                target = np.argmax(chp)
            if target is None:
                return None
            return halo[self._bhtype][target]
        else:
            return halo[self._bhtype]

class BHCentral(BHGal):
    names = 'bhcenter'

    def __init__(self, simulation=None):
        super(BHCentral, self).__init__(simulation, 'BH_central_distance', 'min', 'BH_central')



from __future__ import absolute_import
import numpy as np

import tangos
from tangos.util import consistent_collection
from .. import BuiltinFunction, FixedInput, FixedNumericInput, StoredProperty
from ... import core
from ...core import extraction_patterns

@BuiltinFunction.register
def match(source_halos, target):
    timestep = consistent_collection.ConsistentCollection(source_halos).timestep
    if target is None:
        results = [None]*len(source_halos)
    else:
        from ... import relation_finding
        if not isinstance(target, core.Base):
            target = tangos.get_item(target, core.Session.object_session(timestep))
        results = relation_finding.MultiSourceMultiHopStrategy(source_halos, target).all()
    assert len(results) == len(source_halos)
    return np.array(results, dtype=object)
match.set_input_options(0, provide_proxy=True, assert_class = FixedInput)

@BuiltinFunction.register
def later(source_halos, num_steps):
    timestep = consistent_collection.ConsistentCollection(source_halos).timestep.get_next(num_steps)
    return match(source_halos, timestep)

later.set_input_options(0, provide_proxy=True, assert_class = FixedNumericInput)


@BuiltinFunction.register
def earlier(source_halos, num_steps):
    return later(source_halos, -num_steps)

earlier.set_input_options(0, provide_proxy=True, assert_class = FixedNumericInput)


@BuiltinFunction.register
def latest(source_halos):
    timestep = consistent_collection.ConsistentCollection(source_halos).timestep.get_final()
    return match(source_halos, timestep)

@BuiltinFunction.register
def earliest(source_halos):
    timestep = consistent_collection.ConsistentCollection(source_halos).timestep.get_final(-1)
    return match(source_halos, timestep)

@BuiltinFunction.register
def has_property(source_halos, property):
    return np.not_equal(property,None)

has_property.set_input_options(0, provide_proxy=False, assert_class=StoredProperty)


@BuiltinFunction.register
def has_link(source_halos, link):
    return np.not_equal(link,None)

@has_link.set_initialisation
def link_exists_initialisation(input):

    input.set_extraction_pattern(
        extraction_patterns.halo_link_target_getter)

from . import arithmetic, array, reassembly


from pyramid.view import view_config
from ...core import dictionary
from ... import properties, live_calculation
@view_config(route_name='autocomplete_words', renderer='json')
def autocomplete_words(request):
    dictionary_items = list(dictionary.get_lexicon(request.dbsession))
    liveproperties = properties.all_properties(with_particle_data=False)+live_calculation.BuiltinFunction.all()
    liveproperties = list(map(lambda x: x+"(", liveproperties))
    return dictionary_items+liveproperties



from __future__ import absolute_import
from . import core
from sqlalchemy import Column, Table, String, Integer, Float, ForeignKey
import sqlalchemy
import random
import string
import contextlib
from six.moves import range

_temp_sessions = {}

def _create_temp_halolist(session):
    global _temp_sessions

    rstr = ''.join(random.choice(string.ascii_lowercase) for _ in range(10))
    halolist_table = Table(
            'halolist_'+rstr,
            core.Base.metadata,
            Column('id',Integer, primary_key=True),
            Column('halo_id',Integer,ForeignKey('halos.id')),
            prefixes = ['TEMPORARY']
        )

    halolist_table.create(bind=session.connection())
    _temp_sessions[id(halolist_table)] = session
    return halolist_table

def _delete_temp_halolist(table):
    global _temp_sessions
    connection = _get_connection_for(table)
    table.drop(bind=connection)
    core.Base.metadata.remove(table)
    del _temp_sessions[id(table)]

def _insert_into_temp_halolist(table, ids):
    connection = _get_connection_for(table)
    if isinstance(ids, sqlalchemy.orm.query.Query):
        connection.execute(table.insert().from_select(['halo_id'], ids))
    else:
        connection.execute(
            table.insert(),
            *[{'halo_id': id} for id in ids]
        )

def _get_session_for(table):
    global _temp_sessions
    return _temp_sessions[id(table)]

def _get_connection_for(table):
    global _temp_sessions
    return _temp_sessions[id(table)].connection()

def halo_query(table):
    session = _get_session_for(table)
    return session.query(core.halo.Halo).select_from(table).join(core.halo.Halo)

def all_halos_with_duplicates(table):
    session = _get_session_for(table)
    return [x[1] for x in session.query(table.c.id, core.halo.Halo).select_from(table).outerjoin(
        core.halo.Halo).all()]

def halolink_query(table):
    session = _get_session_for(table)
    return session.query(core.halo_data.HaloLink).select_from(table).join(core.halo_data.HaloLink, core.halo_data.HaloLink.halo_from_id == table.c.halo_id)

@contextlib.contextmanager
def temporary_halolist_table(session, ids=None, callback=None):

    table = _create_temp_halolist(session)
    if ids is not None:
        _insert_into_temp_halolist(table, ids)
    yield table
    _delete_temp_halolist(table)
    if callback is not None:
        callback()



from __future__ import absolute_import
install_requires = [
    'setuptools',
    'numpy >= 1.10.0',
    'sqlalchemy >= 1.0.10',
    'pyparsing >= 2.1.0',
    'WebOb >= 1.7.0rc2', # Response.has_body
    'repoze.lru >= 0.4', # py3 compat
    'zope.interface >= 3.8.0',  # has zope.interface.registry
    'zope.deprecation >= 3.5.0', # py3 compat
    'venusian >= 1.0a3', # ``ignore``
    'translationstring >= 0.4', # py3 compat
    'PasteDeploy >= 1.5.0', # py3 compat
    'plaster',
    'plaster_pastedeploy',
    'pyramid',
    'waitress',
    'pyramid_debugtoolbar',
    'pyramid_jinja2',
    'pyramid_tm','pyramid_retry',
    'zope.sqlalchemy',
    'hupper',
    'six',
    ]

tests_require = [
    'nose >= 1.3.0',
    'webtest >= 2.0',
    'pynbody >= 0.42',
    'yt>=3.4.0'
    ]

from setuptools import setup, find_packages


setup(name='tangos',
      version='1.0.0',
      description='TANGOS, the amazing numerical galaxy organisation system',
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
      ],
      author="Andrew Pontzen",
      author_email="a.pontzen@ucl.ac.uk",
      license="GNUv3",
      packages=find_packages(),
      entry_points={'paste.app_factory': [
                                            'main = tangos.web:main',
                                            ],
                    'console_scripts': [   'tangos_add_bh = tangos.scripts.add_bh:main',
                                           'tangos_bh_timelink = tangos.scripts.bh_timelink:main',
                                           'tangos_crosslink = tangos.scripts.crosslink:main',
                                           'tangos_fix_bh_hosts = tangos.scripts.fix_bh_hosts:main',
                                           'tangos_import_from_ahf = tangos.scripts.import_from_ahf:main',
                                           'tangos_import_from_subfind = tangos.scripts.import_from_subfind:main',
                                           'tangos_manager = tangos.scripts.manager:main',
                                           'tangos_preprocess_bh = tangos.scripts.preprocess_bh:main',
                                           'tangos_timelink = tangos.scripts.timelink:main',
                                           'tangos_writer = tangos.scripts.writer:main'
                    ]
      },
      include_package_data=True,
      zip_safe=False,
      python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      install_requires=install_requires,
      tests_require=tests_require,
      test_suite="nose.collector"
      )



from __future__ import absolute_import
from . import message
from six.moves import range

awaiting_barrier = None


class MessageBarrierPass(message.Message):
    pass

class MessageBarrier(message.Message):
    def process(self):
        from . import backend
        global awaiting_barrier
        if awaiting_barrier is None:
            awaiting_barrier = [False for i in range(backend.size())]

        awaiting_barrier[self.source] = True
        if all(awaiting_barrier[1:]):
            for i in range(1, backend.size()):
                MessageBarrierPass().send(i)
                awaiting_barrier = [False for i in range(backend.size())]


def barrier():
    from . import backend
    assert backend.rank()!=0, "The server process cannot take part in a barrier"
    MessageBarrier().send(0)
    MessageBarrierPass.receive(0)



from __future__ import absolute_import
from pyramid.view import view_config
import pyramid.httpexceptions as exc
import tangos
from . import halo_from_request

def halo_later_earlier(request, rel='later'):
    halo = halo_from_request(request)

    if request.matchdict['n']=='inf':
        rel = rel[:-1]+"st" # later->latest, earlier->earliest
        try:
            halo = halo.calculate("%s()"%(rel))
        except tangos.live_calculation.NoResultsError:
            pass

    else:
        try:
            steps = int(request.matchdict['n'])

            if steps==0:
                return

            steps = str(steps)
        except ValueError:
            steps = "'%s'"%request.matchdict['n']

        try:
            halo = halo.calculate("%s(%s)"%(rel,steps))
        except tangos.live_calculation.NoResultsError:
            pass

    raise exc.HTTPFound(request.route_url("halo_view", simid=halo.timestep.simulation.basename,
                                          timestepid=halo.timestep.extension,
                                          halonumber=halo.basename))



@view_config(route_name='halo_later', renderer=None)
def halo_later(request):
    return halo_later_earlier(request, 'later')


@view_config(route_name='halo_earlier', renderer=None)
def halo_earlier(request):
    return halo_later_earlier(request, 'earlier')

@view_config(route_name='halo_in', renderer=None)
def halo_in(request):
    return halo_later_earlier(request, 'match')



from __future__ import absolute_import
from . import message
from .. import log

j = -1
num_jobs = None
current_job = None

class MessageStartIteration(message.Message):
    def process(self):
        global num_jobs, current_job
        if num_jobs is None:
            num_jobs = self.contents
            current_job = 0
        else:
            if num_jobs != self.contents:
                raise RuntimeError("Number of jobs (%d) expected by rank %d is inconsistent with %d" % (
                self.contents, self.source, num_jobs))


class MessageDeliverJob(message.Message):
    pass

class MessageRequestJob(message.Message):
    def process(self):
        global j, num_jobs, current_job
        source = self.source
        if current_job is not None and num_jobs>0:
            log.logger.info("Send job %d of %d to node %d", current_job, num_jobs, source)
        else:
            current_job = None # in case num_jobs=0, still want to send 'end of loop' signal to client
            log.logger.info("Finished jobs; notify node %d", source)

        MessageDeliverJob(current_job).send(source)

        if current_job is not None:
            current_job += 1
            if current_job == num_jobs:
                num_jobs = None
                current_job = None


def parallel_iterate(task_list):
    """Sets up an iterator returning items of task_list. """
    from . import backend, barrier

    assert backend is not None, "Parallelism is not initialised"

    MessageStartIteration(len(task_list)).send(0)
    barrier()

    while True:
        MessageRequestJob().send(0)
        job = MessageDeliverJob.receive(0).contents

        if job is None:
            barrier()
            return
        else:
            yield task_list[job]


from __future__ import absolute_import
from pyramid.view import view_config
from sqlalchemy import func

from . import simulation_from_request
import socket
import tangos
from tangos import core

@view_config(route_name='simulation_view', renderer='../templates/simulation_view.jinja2')
def simulation_view(request):
    session = request.dbsession

    sim = simulation_from_request(request)

    assert sim is not None

    timesteps = session.query(tangos.core.TimeStep).filter_by(simulation_id=sim.id).\
        order_by(tangos.core.timestep.TimeStep.time_gyr.desc()).all()
    counts = session.query(func.count(core.Halo.id)).\
        join(core.TimeStep).\
        filter(core.TimeStep.simulation_id==sim.id).\
        group_by(core.TimeStep.id). \
        order_by(tangos.core.timestep.TimeStep.time_gyr.desc()).\
        all()

    timestep_links = [request.route_url('timestep_view',simid=sim.basename,timestepid=timestep.extension)
                      for timestep in timesteps]

    counts = [c[0] for c in counts]

    simname = sim.basename

    props = []
    for q in sim.properties:
        props.append((q.name.text, q.data_repr()))

    return {'simulation':simname,
            'timesteps':timesteps,
            'links':timestep_links,
            'counts':counts,
            'properties':props
            }




"""Logic for getting halo properties and links from a Halo object.

Commonly-used patterns are effectively static (i.e. they have no options), and default instances for these patterns
are provided.

e.g.
 - getting a HaloProperty (HaloPropertyGetter -> default instance halo_property_getter);
 - getting a HaloProperty's value (HaloPropertyValueGetter -> default instance halo_property_value_getter);
 - getting a HaloProperty's value without calling reassemble (HaloPropertyRawValueGetter -> halo_property_raw_value_getter);
 - getting a HaloLink (HaloLinkGetter -> halo_link_getter) ;
 - getting a HaloLink target (HaloLinkTargetGetter -> halo_link_target_getter).

However, in one case -- where one wants to pass options to the data reassembly process -- a static implementation is
not possible. So the HaloPropertyValueWithReassemblyOptionsGetter has no "default" instance and has to be
instantiated when required.
"""

from __future__ import absolute_import
import sqlalchemy




class HaloPropertyGetter(object):
    def use_fixed_cache(self, halo):
        return 'all_properties' not in sqlalchemy.inspect(halo).unloaded

    def get(self, halo, property_id, session):
        """Get the specified property, from the in-memory cache if it exists otherwise from the database
        using the specified session

        :type halo: Halo
        :type property_id: int
        :type session: sqlalchemy.orm.session.Session
        """
        if self.use_fixed_cache(halo):
            return self.get_from_cache(halo, property_id)
        else:
            return self.get_from_session(halo, property_id, session)

    def keys(self, halo, session):
        """Get a list of keys, from the in-memory cache if it exists otherwise from the database
            using the specified session

        :type halo: Halo
        :type session: sqlalchemy.orm.session.Session
        """
        if self.use_fixed_cache(halo):
            return self.keys_from_cache(halo)
        else:
            return self.keys_from_session(halo, session)

    def get_from_cache(self, halo, property_id):
        """Get the specified property from an existing in-memory cache

        :type halo: Halo
        :type property_id: int"""

        return_vals = []

        for x in halo.all_properties:
            if x.name_id == property_id:
                return_vals.append(x)

        return self._postprocess(return_vals)


    def get_from_session(self, halo, property_id, session):
        """Get the specified property from the database using the specified session

        :type halo: Halo
        :type property_id: int
        :type session: sqlalchemy.orm.session.Session"""
        from . import halo_data
        query_properties = session.query(halo_data.HaloProperty).filter_by(name_id=property_id, halo_id=halo.id,
                                                                                        deprecated=False).order_by(
            halo_data.HaloProperty.id.desc())

        return self._postprocess(query_properties.all())

    def keys_from_cache(self, halo):
        """Return a list of keys from an existing in-memory cache"""
        return [x.name.text for x in halo.all_properties]

    def keys_from_session(self, halo, session):
        from . import halo_data
        query_properties = session.query(halo_data.HaloProperty).filter_by(halo_id=halo.id,
                                                                                        deprecated=False)
        return [x.name.text for x in query_properties.all()]

    def cache_contains(self, halo, property_id):
        """Return True if the existing in-memory cache has the specified property

        :type halo: Halo
        :type property_id: int"""

        for x in halo.all_properties:
            if x.name_id == property_id:
                return True

        return False


    def _postprocess(self, outputs):
        return outputs

halo_property_getter = HaloPropertyGetter()

class HaloPropertyValueGetter(HaloPropertyGetter):
    def _postprocess(self, outputs):
        return [o.get_data_with_reassembly_options() for o in outputs]
    
halo_property_value_getter = HaloPropertyValueGetter()


class HaloPropertyValueWithReassemblyOptionsGetter(HaloPropertyGetter):
    def __init__(self, *options):
        self._options = options

    def _postprocess(self, outputs):
        return [o.get_data_with_reassembly_options(*self._options) for o in outputs]


class HaloPropertyRawValueGetter(HaloPropertyGetter):
    def _postprocess(self, outputs):
        return [o.data_raw for o in outputs]

halo_property_raw_value_getter=HaloPropertyRawValueGetter()


class HaloLinkGetter(HaloPropertyGetter):
    def get_from_cache(self, halo, property_id):
        return_vals = []

        for x in halo.all_links:
            if x.relation_id == property_id:
                return_vals.append(x)

        return self._postprocess(return_vals)

    def get_from_session(self, halo, property_id, session):
        from . import halo_data
        query_links = session.query(halo_data.HaloLink).filter_by(relation_id=property_id, halo_from_id=halo.id)
        return self._postprocess(query_links.all())

    def cache_contains(self, halo, property_id):
        for x in halo.all_links:
            if x.relation_id == property_id:
                return True

        return False

    def keys_from_cache(self, halo):
        """Return a list of keys from an existing in-memory cache"""
        return [x.relation.text for x in halo.all_links]

    def keys_from_session(self, halo, session):
        from . import halo_data
        query_properties = session.query(halo_data.HaloLink).filter_by(halo_from_id=halo.id)
        return [x.relation.text for x in query_properties.all()]
    
halo_link_getter = HaloLinkGetter()


class HaloLinkTargetGetter(HaloLinkGetter):
    def _postprocess(self, outputs):
        return [o.halo_to for o in outputs]

halo_link_target_getter = HaloLinkTargetGetter()


class ReverseHaloLinkGetter(HaloLinkGetter):
    def get_from_cache(self, halo, property_id):
        return_vals = []

        for x in halo.all_reverse_links:
            if x.relation_id == property_id:
                return_vals.append(x)

        return self._postprocess(return_vals)

    def get_from_session(self, halo, property_id, session):
        from . import halo_data
        query_links = session.query(halo_data.HaloLink).filter_by(relation_id=property_id,
                                                                               halo_to_id=halo.id)
        return self._postprocess(query_links.all())

    def cache_contains(self, halo, property_id):
        for x in halo.all_reverse_links:
            if x.relation_id == property_id:
                return True

        return False


class ReverseHaloLinkSourceGetter(ReverseHaloLinkGetter):
    def _postprocess(self, outputs):
        return [o.halo_from for o in outputs]


from . import PynbodyHaloProperties

class MeanGasProperties(PynbodyHaloProperties):
    @classmethod
    def name(self):
        return "mean_temp", "mean_rho"

    def calculate(self, particle_data, properties):
        return particle_data.gas['temp'].mean(), particle_data.gas['rho'].mean().in_units("m_p cm^-3")



import socket

class BreadCrumbs(object):
    def __init__(self, uri=None, servername="TANGOS"):
        if uri:
            self.uri = uri  # invoke uri.setter
        else:
            self._uri = uri  # set _uri to none
        self.servername = servername

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        """set the _uri, _protocol, and crumbs attributes"""
        self._protocol = 'http://'
        protocols = ['http://', 'https://', 'ftp://', 'sftp://']
        for protocol in protocols:
            if uri.startswith(protocol):
                self._protocol = protocol
                uri = uri[len(protocol):]  # remove protocol from uri

        self._uri = uri.rstrip('/')
        self.crumbs = self._uri.split('/')

    @property
    def links(self):
        links = []
        for count, crumb in enumerate(self.crumbs, start=1):
            crumb_uri = self._protocol + '/'.join(self.crumbs[0:count])
            if count==1:
                crumb = self.servername
            if count<len(self.crumbs):
                links.append('<a href="' + crumb_uri + '" class="breadcrumbs">' + crumb + '</a>')
            else:
                links.append(crumb)
        return links

def breadcrumbs(request):
    c = BreadCrumbs(request.url, servername="TANGOS on "+socket.gethostname())
    return " &rarr; ".join(c.links)


from __future__ import absolute_import
import contextlib
import time
import inspect
import numpy as np
import six

class TimingMonitor(object):
    """This class keeps track of how long a Property is taking to evaluate, and (if the Property is implemented
    to take advantage of this), the time spent on sub-tasks. It provides formatting to place this information
    into the log."""
    def __init__(self):
        self.timings_by_class = {}
        self.labels_by_class = {}
        self._monitoring = None

    @contextlib.contextmanager
    def __call__(self, object):
        self._start(object)
        yield
        self._end()

    def check_compatible_object(self, object):
        if not hasattr(object, 'timing_monitor'):
            raise TypeError("TimingMonitor requires a compatible object to monitor")

    def _set_as_monitor_for(self, object):
        self.check_compatible_object(object)
        if self._monitoring is not None:
            raise RuntimeError(
                "TimingMonitor is already monitoring %r and cannot also monitor %r" % (self._monitoring, object))
        self._monitoring = object
        self._old_timing_monitor = object.timing_monitor
        object.timing_monitor = self

    def _unset_as_monitor_for(self, object):
        if self._monitoring is not object:
            raise RuntimeError("Consistency error - trying to stop monitoring an object that is not being monitored")
        object.timing_monitor = self._old_timing_monitor
        del self._old_timing_monitor
        self._monitoring = None

    def _start(self, object):
        """Start a timer for the specified object"""
        self._set_as_monitor_for(object)
        self._time_marks_info = ["start"]
        self._time_marks = [time.time()]

    def _end(self):
        """End a timer for the specified object."""
        cl = type(self._monitoring)
        self._unset_as_monitor_for(self._monitoring)
        self._time_marks_info.append("end")
        self._time_marks.append(time.time())
        previous_timings = self.timings_by_class.get(cl, 0)
        cumulative_timings = np.diff(self._time_marks)+previous_timings
        self.timings_by_class[cl] = cumulative_timings
        self.labels_by_class[cl] = self._time_marks_info


    def mark(self, label=None):
        """Mark a named event so that more detailed timing can be given"""
        self._time_marks.append(time.time())
        if label is None:
            self._time_marks_info.append(
                inspect.currentframe().f_back.f_lineno)
        else:
            self._time_marks_info.append(label)

    def summarise_timing(self, logger):
        logger.info("CUMULATIVE RUNNING TIMES (just this node)")
        v_tot = 1e-10
        for k, v in six.iteritems(self.timings_by_class):
            v_tot += sum(v)

        for k, v in six.iteritems(self.timings_by_class):
            name = str(k)[:-2]
            name = name.split(".")[-1]
            name = "%20s " % (name[-20:])
            if len(v)>1:
                marks_info = self.labels_by_class[k]
                logger.info(" " + name + "%.1fs | %.1f%%" % (sum(v), 100 * sum(v) / v_tot))
                logger.info("  ------ INTERNAL BREAKDOWN ------" )
                for i, this_v in enumerate(v):
                    logger.info((" %8s %8s %.1fs | %.1f%% | %.1f%%") %
                                (marks_info[i], marks_info[i + 1],
                                 this_v, 100 * this_v / sum(v), 100 * this_v / v_tot))
                logger.info("  --------------------------------")
            else:
                logger.info(name + "%.1fs | %.1f%%" % (v[0], 100 * v[0] / v_tot))




"""The live calculation system, used by Halo.calculate, Halo.calculate_for_descendants and TimeStep.calculate_all.

For more overview information, see live_calculation.md. """

from __future__ import absolute_import
import warnings

import numpy as np
from sqlalchemy.orm import contains_eager, aliased

import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data
from tangos.core import extraction_patterns
from tangos.live_calculation.query_masking import QueryMask
from tangos.util import consistent_collection
from .. import core
from .. import temporary_halolist as thl
from six.moves import range
from six.moves import zip

class UnknownValue(object):
    """A dummy object returned by Calculation.proxy_value when the value of the calculation cannot be predicted"""
    pass

class NoResultsError(ValueError):
    pass

class Calculation(object):
    """Represents a live-calculation to be performed on database objects.

    A typical Calculation is composed of many different Calculation subclasses arranged in some form of tree.
    For example A.B parses to Link(StoredProperty('A'), StoredProperty('B')) whereas
                A.B() parses to Link(StoredProperty('A'), LiveProperty('B'))
    and so on.

    It is normally easiest to instantiate Calculation objects using the parser in
    live_calculation.parser"""

    def __init__(self):
        self._extraction_pattern = extraction_patterns.halo_property_value_getter

    def __repr__(self):
        return "<Calculation description for %s>"%str(self)

    def __str__(self):
        raise NotImplementedError

    def set_extraction_pattern(self, extraction_pattern):
        self._extraction_pattern = extraction_pattern

    def retrieves(self):
        """Return the set of named halo properties that this calculation will access

        Redirections are indicated with a ".", e.g. if a calculation access the 'mass' property of a halo linked
        by 'BH', it will include 'BH.mass' in the set."""
        return set()

    def name(self):
        """Returns the name of this calculation, formatted such that parser.parse_property_name(name) generates a copy."""
        return None

    def _has_required_properties(self, halo):
        property_is_present = []
        for p_id in self._essential_dict_ids():
            this_property_ok = False
            for extraction_pattern in (extraction_patterns.halo_link_getter,
                                       extraction_patterns.halo_property_getter):
                if not extraction_pattern.use_fixed_cache(halo):
                    this_property_ok = True
                elif extraction_pattern.cache_contains(halo, p_id):
                    this_property_ok = True
            property_is_present.append(this_property_ok)
        return all(property_is_present)

    def retrieves_dict_ids(self):
        """Returns the dictionary IDs of the named properties to be retrieved for each halo to
        allow this calculation to run"""
        self._generate_dict_ids_and_levels()
        return self._r_dict_ids_cached

    def _essential_dict_ids(self):
        """Returns the dictionary IDs of properties that must be present in a halo for the calculation to run.

        If any of the required IDs are not present, the result of the calculation is assumed to be None"""
        self._generate_dict_ids_and_levels()
        return self._r_dict_ids_essential_cached

    def n_join_levels(self):
        """Return the number of levels of HaloLinks that evaluating this property requires.

        For example, evaluating 'mass' has n_join_levels=0; 'BH.mass' has n_join_levels=1;
        'BH.other.mass' has n_join_levels=2"""
        self._generate_dict_ids_and_levels()
        return self._n_join_levels

    def _generate_dict_ids_and_levels(self):
        if not hasattr(self, "_r_dict_ids_cached"):
            session = core.Session()
            try:
                self._r_dict_ids_cached = set()
                self._r_dict_ids_essential_cached = set()
                retrieves = self.retrieves()
                try:
                    self._n_join_levels = max([r.count('.') for r in retrieves])+1
                except ValueError:
                    self._n_join_levels = 0
                for r in retrieves:
                    r_split = r.split(".")
                    for w in r_split:
                        w_dict_id = tangos.core.dictionary.get_dict_id(w, -1, session=session)
                        if w_dict_id!=-1:
                            self._r_dict_ids_cached.add(w_dict_id)
                    r_dict_id = tangos.core.dictionary.get_dict_id(r_split[0], -1, session=session)
                    if r_dict_id!=-1:
                        self._r_dict_ids_essential_cached.add(r_dict_id)
            finally:
                session.close()

    def values_and_description(self, halos):
        """Return the values of this calculation, as well as a HaloProperties object describing the
        properties of these values (if possible)"""
        raise NotImplementedError

    def values_sanitized_and_description(self, halos):
        """Return the 'sanitized' values of this calculation, as well as a HaloProperties object (if available).

        See values_sanitized for the definition of sanitized"""
        values, desc = self.values_and_description(halos)
        return self._sanitize_values(values), desc

    def values(self, halos):
        """Return the values of this calculation applied to halos.

        The size of the returned numpy object array is self.n_columns() x len(halos) """
        values, _ = self.values_and_description(halos)
        return values

    def value(self, halo):
        """Return the value of this calculation applied to the given halo"""
        return self.values([halo])[:,0]

    def value_sanitized(self, halo):
        """"Return the value of this calculation applied to the given halo, with conversion to numpy arrays where possible.

        See values_sanitized for information about the conversion."""
        return self.values_sanitized([halo])[:,0]

    def values_sanitized(self, halos):
        """Return the values of this calculation applied to halos, with conversion to numpy arrays where possible

        The return value is a self.n_columns()-length list, with each entry in the list being a numpy array of
        length len(halos). The dtype of the numpy array is chosen per-property to match the dtype result found
        when evaluating the first halo."""
        unsanitized_values = self.values(halos)
        return self._sanitize_values(unsanitized_values)



    def _sanitize_values(self, unsanitized_values):
        """Convert the raw calculation result to a sanitized version. See values_sanitized."""
        keep_rows = np.all([[data is not None for data in row] for row in unsanitized_values], axis=0)
        # The obvious way of doing this:
        #   keep_rows = np.all(np.not_equal(out,None), axis=0)
        # generates a scary FutureWarning that it will stop working in future. This is because
        # it can end up doing a comparison of an array to None (effectively data!=None),
        # which will not be the same as "is not None" in future versions of numpy.

        unsanitized_values = unsanitized_values[:, keep_rows]
        for x in unsanitized_values:
            if len(x)==0:
                raise NoResultsError("Calculation %s returned no results" % self)
        return [self._make_numpy_array(x) for x in unsanitized_values]

    @staticmethod
    def _make_numpy_array(x):
        if isinstance(x[0], np.ndarray):
            try:
                return np.array(list(x), dtype=x[0].dtype)
            except ValueError:
                return x
        else:
            return np.array(x, dtype=type(x[0]))

    def n_columns(self):
        """Return the number of separate properties returned for each halo.

        The return from values_sanitized has shape n_columns() x n_halos"""
        return 1

    def supplement_halo_query(self, halo_query):
        """Return a sqlalchemy query with a supplemental join to allow this calculation to run efficiently"""
        name_targets = self.retrieves_dict_ids()
        halo_alias = tangos.core.halo.Halo
        augmented_query = halo_query
        for i in range(self.n_join_levels()):
            halo_property_alias = aliased(tangos.core.halo_data.HaloProperty)
            halo_link_alias = aliased(tangos.core.halo_data.HaloLink)

            path_to_properties = [tangos.core.halo.Halo.all_links, tangos.core.halo_data.HaloLink.halo_to] * i + [
                tangos.core.halo.Halo.all_properties]
            path_to_links = [tangos.core.halo.Halo.all_links, tangos.core.halo_data.HaloLink.halo_to] * i + [
                tangos.core.halo.Halo.all_links]


            if len(name_targets)>0:
                property_name_condition = halo_property_alias.name_id.in_(name_targets)
                link_name_condition = (halo_link_alias.relation_id.in_(name_targets))
            else:
                # We know we're joining to a null list of properties; do this as efficiently as possible
                property_name_condition = link_name_condition = False


            augmented_query =augmented_query.outerjoin(halo_property_alias,
                                                  (halo_alias.id==halo_property_alias.halo_id)
                                                  & property_name_condition).\
                                        outerjoin(halo_link_alias,
                                                  (halo_alias.id==halo_link_alias.halo_from_id)
                                                  & link_name_condition).\
                                        options(contains_eager(*path_to_properties, alias=halo_property_alias),
                                                contains_eager(*path_to_links, alias=halo_link_alias))

            if i<self.n_join_levels()-1:
                next_level_halo_alias = aliased(tangos.core.halo.Halo)
                path_to_new_halo = path_to_links + [tangos.core.halo_data.HaloLink.halo_to]
                augmented_query = augmented_query.outerjoin(next_level_halo_alias,
                                                            (halo_link_alias.halo_to_id==next_level_halo_alias.id)).\
                                        options(contains_eager(*path_to_new_halo, alias=next_level_halo_alias))

                halo_alias = next_level_halo_alias

        return augmented_query

    def proxy_value(self):
        """Return a placeholder value for this calculation"""
        raise NotImplementedError


class MultiCalculation(Calculation):
    """Represents a single calculation that returns the results from multiple sub-calculations."""
    def __init__(self, *calculations):
        super(MultiCalculation, self).__init__()
        self.calculations = [c if isinstance(c, Calculation) else parser.parse_property_name(c) for c in calculations]

    def retrieves(self):
        x = set()
        for c in self.calculations:
            x.update(c.retrieves())
        return x

    def __str__(self):
        return "("+(", ".join(str(x) for x in self.calculations))+")"

    def values_and_description(self, halos):
        results = np.empty((self.n_columns(),len(halos)), dtype=object)
        c_column = 0
        halos = np.asarray(halos, dtype=object)
        mask = QueryMask()
        mask.mark_nones_as_masked(halos)
        for c in self.calculations:
            values, description = c.values_and_description(mask.mask(halos))
            results[c_column:c_column+c.n_columns()] = mask.unmask(values)
            # TODO: in principle this masking should _not_ occur unless we know the user has called values_sanitized
            # - other calls should not cross-contaminate columns in this way
            mask.mark_nones_as_masked(values)
            c_column+=c.n_columns()

        # TODO - problem: there is no good description of multiple properties
        return results, description

    def n_columns(self):
        return sum(c.n_columns() for c in self.calculations)


class FixedInput(Calculation):
    """Represents a calculation that returns a fixed value"""
    def __init__(self, *tokens):
        super(FixedInput, self).__init__()
        self.value = str(tokens[0])

    def __str__(self):
        return '"'+str(self.value)+'"'

    def values(self, halos):
        return np.array([[self.value]*len(halos)],dtype=object)

    def values_and_description(self, halos):
        return self.values(halos), self.value

    def proxy_value(self):
        return self.value

class FixedNumericInput(FixedInput):
    """Represents a calculation that returns a fixed numerical value"""
    def __init__(self, *tokens):
        super(FixedNumericInput, self).__init__()

    @staticmethod
    def _process_numerical_value(t):
        if "." in t or "e" in t or "E" in t:
            return float(t)
        else:
            return int(t)

    def __init__(self, *tokens):
        self.value = self._process_numerical_value(tokens[0])

    def __str__(self):
        return str(self.value)

class LiveProperty(Calculation):
    """Represents a calculation that is achieved by executing the live_calculate method of a Properties instance"""
    def __new__(cls, *tokens):
        if BuiltinFunction.has_function(str(tokens[0])):
            return object.__new__(BuiltinFunction)
        else:
            return object.__new__(LiveProperty)


    def __init__(self, *tokens):
        super(LiveProperty, self).__init__()
        self._name = str(tokens[0])
        self._inputs = list(tokens[1:])

    def __str__(self):
        return self._name + "(" + (",".join(str(x) for x in self._inputs)) + ")"

    def name(self):
        return self._name

    def retrieves(self):
        result = self._calculation_retrieves()
        result = result.union(self._parameters_retrieve())
        return result

    def _parameters_retrieve(self):
        result = set()
        for i in self._inputs:
            result = result.union(i.retrieves())
        return result

    def _calculation_retrieves(self):
        from .. import properties
        result = set()
        proxy_values = [i.proxy_value() for i in self._inputs]
        providing_instance = properties.providing_class(self._name)(None, *proxy_values)
        result = result.union(providing_instance.requires_property())
        return result

    def values_and_description(self, halos):
        input_values = []
        input_descriptions = []
        for input in range(len(self._inputs)):
            iv, id = self._input_value_and_description(input, halos)
            input_values.append(iv)
            input_descriptions.append(id)

        calculator, results = self._evaluate_function(halos, input_descriptions, input_values)

        return results, calculator

    def _input_value_and_description(self, input_id, halos):
        input = self._inputs[input_id]
        val, desc = input.values_and_description(halos)
        if len(val) != 1:
            raise ValueError("Functions cannot receive more than one value per input")
        return val[0], desc

    def _evaluate_function(self, halos, input_descriptions, input_values):
        from .. import properties
        sim = consistent_collection.consistent_simulation_from_halos(halos)
        results = []
        calculator = properties.providing_class(self.name())(sim, *input_descriptions)
        for inputs in zip(halos, *input_values):
            if self._has_required_properties(inputs[0]) and all([x is not None for x in inputs]):
                results.append(calculator.live_calculate_named(self.name(), *inputs))
            else:
                results.append(None)
        return calculator, self._as_1xn_array(results)

    @classmethod
    def _as_1xn_array(cls, results):
        results_array = np.empty((1, len(results)), dtype=object)
        results_array[0, :] = results
        return results_array

    def proxy_value(self):
        return UnknownValue()

class BuiltinFunction(LiveProperty):
    """Represents a calculation that is achieved by executing a python function. See the builtin_functions module."""

    __registered_functions = {}
    __default_args = {'provide_proxy': False, 'assert_class': None}

    @classmethod
    def all(cls):
        return list(cls.__registered_functions.keys())

    @classmethod
    def register(cls, func):
        """Register a built-in function for the live calculation (LC) system.

        The LC-exposed version of the function inherits the same name as the python function.

        The python function takes a list of halos and then further lists for each argument given to the
        live-calculation. By default, these arugments are lists of the values for each halo. However
        this behaviour can be customised by calling set_input_options.

        For examples, see the builtin_functions module.
        """
        cls.__registered_functions[func.__name__] = {'function': func}
        func.set_input_options = lambda arg_id, **kwargs: cls.set_input_options(func, arg_id, **kwargs)
        func.set_initialisation = lambda init_fn: cls.set_initialisation(func, init_fn)
        return func

    @classmethod
    def set_input_options(cls, func, argument_id, **kwargs):
        """For the registered function, set input options.

        :param argument_id: The 0-based ID of the input argument (as seen in the LC langauge,
                            i.e. discounting the halos input)
        :param provide_proxy: If True, the value of the input is not evaluated and only a proxy is passed
        :param assert_class: If not None, the input must be an instance of the provided class
        """
        for k in kwargs.keys():
            if k not in list(cls.__default_args.keys()):
                raise ValueError("Unknown input option %s"%k)
        cls.__registered_functions[func.__name__][argument_id] = kwargs

    @classmethod
    def set_initialisation(cls, func, initialisation_func):
        """For the registered function, add an initialisation function that receives the input objects"""
        cls.__registered_functions[func.__name__]['initialisation'] = initialisation_func

    @classmethod
    def has_function(cls, func_name):
        return func_name in list(cls.__registered_functions.keys())

    def __init__(self, *tokens):
        super(BuiltinFunction, self).__init__(*tokens)
        self._func = self.__registered_functions[self._name]['function']
        self._info = self.__registered_functions[self._name]
        for i in range(len(self._inputs)):
            assert_class = self._get_input_option(i, 'assert_class')
            if assert_class is not None and not isinstance(self._inputs[i], assert_class):
                raise ValueError("Input %d to function %s must be of type %s (received %s)"%\
                                  (i,self._name,assert_class,type(self._inputs[i])))
        if 'initialisation' in self._info:
            self._info['initialisation'](*self._inputs)

    def _get_input_option(self, input_id, option):
        default = self.__default_args[option]
        if input_id in self._info:
            return self._info[input_id].get(option, default)
        else:
            return default

    def _input_value_and_description(self, input_id, halos):
        if self._get_input_option(input_id, 'provide_proxy'):
            return self._inputs[input_id].proxy_value(), None
        else:
            return super(BuiltinFunction, self)._input_value_and_description(input_id, halos)

    def _calculation_retrieves(self):
        return set()

    def _evaluate_function(self, halos, input_descriptions, input_values):
        if len(input_descriptions)>0:
            inherited_description = input_descriptions[0]
        else:
            inherited_description = None
        return inherited_description, self._as_1xn_array(self._func(halos, *input_values))



class Link(Calculation):
    """Represents a calculation to be made on a halo linked to the input in some way"""
    def __init__(self, *tokens):
        super(Link, self).__init__()
        self.locator = tokens[0]
        self.property = tokens[1]
        if not isinstance(self.locator, Calculation):
            self.locator = parser.parse_property_name(self.locator)
        if not isinstance(self.property, Calculation):
            self.property = parser.parse_property_name(self.property)
        self.locator.set_extraction_pattern(extraction_patterns.halo_link_target_getter)

    def __str__(self):
        return str(self.locator)+"."+str(self.property)

    def name(self):
        return self.property.name()

    def proxy_value(self):
        """Return a placeholder value for this calculation"""
        return UnknownValue()

    def retrieves(self):
        # the property retrieval will not be on the set of halos known to higher levels,
        # so only the locator retrieval needs to be reported upwards
        return self.locator.retrieves()

    def n_columns(self):
        return self.property.n_columns()

    def values_and_description(self, halos):
        if self.locator.n_columns()!=1:
            raise ValueError("Cannot use property %r, which returns more than one column, as a halo locator"%(str(self.locator)))

        target_halos = self._get_target_halos(halos)
        results = np.empty((self.n_columns(),len(halos)),dtype=object)

        mask = QueryMask()
        mask.mark_nones_as_masked(target_halos)
        target_halo_masked = mask.mask(target_halos)


        for i in range(len(target_halo_masked)):
            if isinstance(target_halo_masked[i], list):
                warnings.warn("More than one relation for target %r has been found. Picking the first."%str(self.locator))
                target_halo_masked[i] = target_halo_masked[i][0].id
            else:
                target_halo_masked[i] = target_halo_masked[i].id

        # need a new session for the subqueries, because we might have cached copies of objects where
        # a different set of properties has been loaded into all_properties
        new_session = core.Session()

        try:
            with thl.temporary_halolist_table(new_session, target_halo_masked) as tab:
                target_halos_supplemented = self.property.supplement_halo_query(thl.halo_query(tab)).all()

                # sqlalchemy's deduplication means we are now missing any halos that appear more than once in
                # target_halos_ids_weeded. But we actually want the duplication.
                target_halos_supplemented_with_duplicates = \
                    self._add_entries_for_duplicates(target_halos_supplemented, target_halo_masked)

                values, description = self.property.values_and_description(target_halos_supplemented_with_duplicates)
        finally:
            new_session.connection().close()

        results[:] = mask.unmask(values)

        return results, description

    def _get_target_halos(self, source_halos):
        target_halos = self.locator.values(source_halos)[0]
        return target_halos

    @staticmethod
    def _add_entries_for_duplicates(target_objs, target_ids):
        if len(target_objs)==len(target_ids):
            return target_objs
        target_obj_ids = [t.id for  t in target_objs]
        return [target_objs[target_obj_ids.index(t_id)] for t_id in target_ids]




class StoredProperty(Calculation):
    """Represents a named property with value stored in the database"""

    def __init__(self, *tokens):
        super(StoredProperty,self).__init__()
        self._name = tokens[0]

    def __str__(self):
        return self._name

    def name(self):
        return self._name

    def retrieves(self):
        return {self._name}

    def values(self, halos):
        self._name_id = tangos.core.dictionary.get_dict_id(self._name)
        ret = np.empty((1,len(halos)),dtype=object)
        for i, h in enumerate(halos):
            if self._extraction_pattern.cache_contains(h, self._name_id):
                ret[0,i]=self._extraction_pattern.get_from_cache(h, self._name_id)[0]
        return ret

    def values_and_description(self, halos):
        from .. import properties
        values = self.values(halos)
        sim = consistent_collection.consistent_simulation_from_halos(halos)
        description_class = properties.providing_class(self._name, sim.output_handler_class, silent_fail=True)
        description = None if description_class is None else description_class(sim)
        return values, description

    def proxy_value(self):
        """Return a placeholder value for this calculation"""
        return UnknownValue()





from . import parser
from . import builtin_functions


from __future__ import absolute_import
import weakref
import os, os.path
import sqlalchemy
from sqlalchemy import Column, Integer, String, ForeignKey, Float, Boolean, and_
from sqlalchemy.orm import relationship, backref, aliased

from . import Base
from .creator import Creator
from .simulation import Simulation
from .. import config

class TimeStep(Base):
    __tablename__ = 'timesteps'

    id = Column(Integer, primary_key=True)
    extension = Column(String)
    simulation_id = Column(Integer, ForeignKey('simulations.id'))
    redshift = Column(Float)
    time_gyr = Column(Float)
    creator = relationship(Creator, backref=backref(
        'timesteps', cascade='delete', lazy='dynamic'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))

    available = Column(Boolean, default=True)

    simulation = relationship(Simulation, backref=backref('timesteps', order_by=time_gyr,
                                                          cascade='delete, merge'),
                              cascade='save-update, merge')

    @property
    def filename(self):
        return os.path.join(config.base, self.simulation.basename, self.extension)

    @property
    def relative_filename(self):
        return str(self.simulation.basename + "/" + self.extension)

    def load(self, mode=None):
        handler = self.simulation.get_output_handler()
        return handler.load_timestep(self.extension, mode=mode)

    def load_region(self, region_specification, mode=None):
        handler = self.simulation.get_output_handler()
        return handler.load_region(self.extension, region_specification, mode=mode)

    def __init__(self, simulation, extension):
        from . import creator
        self.extension = str(extension)
        self.simulation = simulation
        self.creator_id = creator.get_creator_id()

    def __repr__(self):
        extra = ""
        if not self.available:
            extra += " unavailable"
        path = self.path
        if self.redshift is None:
            return "<TimeStep %r%s>"%(path,extra)
        else:
            return "<TimeStep %r z=%.2f t=%.2f Gyr%s>" % (path, self.redshift, self.time_gyr, extra)

    def short(self):
        return "<TimeStep(... z=%.2f ...)>" % self.redshift

    def __getitem__(self, halo_identifier):
        from . import Session, Halo
        session = Session.object_session(self)
        if isinstance(halo_identifier, int):
            object_typecode, halo_number = 0, halo_identifier
        elif "." in halo_identifier:
            object_typecode, halo_number = list(map(int, halo_identifier.split(".")))
        elif "_" in halo_identifier:
            object_typecode, halo_number = halo_identifier.split("_")
            object_typecode = Halo.class_from_tag(object_typecode).__mapper_args__['polymorphic_identity']
            halo_number = int(halo_number)
        else:
            object_typecode, halo_number = 0, int(halo_identifier)
        return session.query(Halo).filter_by(timestep_id=self.id, halo_number=halo_number, object_typecode=object_typecode).first()

    @property
    def path(self):
        return self.simulation.path+"/"+self.extension

    @property
    def redshift_cascade(self):
        a = self.next
        if a is not None:
            return [self.redshift] + a.redshift_cascade
        else:
            return [self.redshift]

    @property
    def time_gyr_cascade(self):
        a = self.next
        if a is not None:
            return [self.time_gyr] + a.time_gyr_cascade
        else:
            return [self.time_gyr]

    def keys(self):
        """Return keys for which ALL halos have a data entry"""
        from . import Session
        session = Session.object_session(self)
        raise RuntimeError("Not implemented")

    def calculate_all(self, *plist, **kwargs):
        """Gather the specified properties from the child objects.

        The parameters passed name the properties (or live-calculations) to return.
        For example m,r = ts.calculate_all("mass","radius") generates an array of mass and
        radius for all objects in timestep ts.

        :param object_type: integer or string representing the particular object type
                            (e.g. 'halo', 'BH' or 'group'). If None (default), all
                            types are included.
        """

        from .. import live_calculation
        from . import Session
        from .halo import Halo

        object_typecode = None
        object_typetag = kwargs.get('object_typetag',None)
        if object_typetag:
            object_typecode = Halo.object_typecode_from_tag(object_typetag)

        if isinstance(plist[0], live_calculation.Calculation):
            property_description = plist[0]
        else:
            property_description = live_calculation.parser.parse_property_names(*plist)

        # must be performed in its own session as we intentionally load in a lot of
        # objects with incomplete lazy-loaded properties
        session = Session()
        try:
            raw_query = session.query(Halo).filter_by(timestep_id=self.id)
            if object_typecode is not None:
                raw_query = raw_query.filter_by(object_typecode=object_typecode)
            query = property_description.supplement_halo_query(raw_query)
            results = query.all()
            results = property_description.values_sanitized(results)
        finally:
            session.close()
        return results

    def gather_property(self, *args, **kwargs):
        """The old alias for calculate_all, retained for compatibility"""
        return self.calculate_all(*args, **kwargs)

    @property
    def earliest(self):
        return self.get_final(-1)

    @property
    def latest(self):
        return self.get_final(+1)

    @property
    def next(self):
        try:
            return self._next
        except AttributeError:
            self._next = self.get_next()
            return self._next

    @property
    def previous(self):
        try:
            return self._previous
        except AttributeError:
            self._previous = self.get_next(-1)
            return self._previous

    def get_next(self, steps=1):
        """Returns the next timestep, or its successor after the specified number of steps.

        If steps is negative, finds the previous timestep or predecessor by specified number of steps.

        If no such step exists, returns None."""

        if steps==0:
            return self
        from . import Session
        session = Session.object_session(self)

        if steps>0:
            direction_comparison = TimeStep.time_gyr > self.time_gyr
            first_order = TimeStep.time_gyr
        else:
            direction_comparison = TimeStep.time_gyr < self.time_gyr
            first_order = TimeStep.time_gyr.desc()

        successors_query = session.query(TimeStep).filter(
               and_(direction_comparison, TimeStep.simulation == self.simulation)
            ).order_by(first_order).limit(abs(steps))

        if successors_query.count()<abs(steps):
            return None

        successors_subquery = aliased(TimeStep, alias=successors_query.subquery())

        if steps>0:
            second_order = successors_subquery.time_gyr.desc()
        else:
            second_order = successors_subquery.time_gyr

        next = session.query(successors_subquery).order_by(second_order).first()

        return next

    def get_final(self, direction=1):
        """Returns the final timestep of this simulation, either the latest (direction=+1) or first (direction=-1)"""

        assert direction==1 or direction==-1

        from . import Session
        session = Session.object_session(self)

        q = session.query(TimeStep).filter_by(simulation_id=self.simulation_id)

        if direction==-1:
            q = q.order_by(TimeStep.time_gyr)
        else:
            q = q.order_by(TimeStep.time_gyr.desc())

        return q.first()





from __future__ import absolute_import
import logging
from six import StringIO
import copy

logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)
handler_stderr = logging.StreamHandler()
handler_stderr.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s : %(message)s")
handler_stderr.setFormatter(formatter)
logger.addHandler(handler_stderr)


class LogCapturer(object):
    def __init__(self):
        self.buffer = StringIO()
        self.handler_buffer = logging.StreamHandler(self.buffer)
        self.handler_buffer.setLevel(logging.INFO)
        self._suspended_handlers = []

    def __enter__(self):
        self._suspended_handlers = copy.copy(logger.handlers)
        for x_handler in self._suspended_handlers:
            logger.removeHandler(x_handler)
        logger.addHandler(self.handler_buffer)

    def __exit__(self, *exc_info):
        for x_handler in self._suspended_handlers:
            logger.addHandler(x_handler)
        self._suspended_handlers = []
        logger.removeHandler(self.handler_buffer)

    def get_output(self):
        return self.buffer.getvalue()


def set_identity_string(identifier):
    global handler_stderr
    formatter = logging.Formatter(identifier+"%(asctime)s : %(message)s")
    handler_stderr.setFormatter(formatter)